@@ -1,40 +1,32 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
+** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
 ** Contact: http://www.qt-project.org/legal
 **
 ** This file is part of the Qt Quick Controls module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL$
+** $QT_BEGIN_LICENSE:LGPL21$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://qt.digia.com/licensing.  For further information
+** a written agreement between you and Digia. For licensing terms and
+** conditions see http://qt.digia.com/licensing. For further information
 ** use the contact form at http://qt.digia.com/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU Lesser General Public License version 2.1 requirements
-** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 2.1 or version 3 as published by the Free
+** Software Foundation and appearing in the file LICENSE.LGPLv21 and
+** LICENSE.LGPLv3 included in the packaging of this file. Please review the
+** following information to ensure the GNU Lesser General Public License
+** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
+** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
 **
 ** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
+** rights. These rights are described in the Digia Qt LGPL Exception
 ** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
 **
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3.0 as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL included in the
-** packaging of this file.  Please review the following information to
-** ensure the GNU General Public License version 3.0 requirements will be
-** met: http://www.gnu.org/copyleft/gpl.html.
-**
-**
 ** $QT_END_LICENSE$
 **
 ****************************************************************************/
@@ -50,7 +42,10 @@
 #include <qsgsimpletexturenode.h>
 #include <qquickwindow.h>
 #include "private/qguiapplication_p.h"
+#include <QtQuick/private/qquickwindow_p.h>
+#include <QtQuick/private/qquickitem_p.h>
 #include <QtGui/qpa/qplatformtheme.h>
+#include "../qquickmenuitem_p.h"
 
 QT_BEGIN_NAMESPACE
 
@@ -95,7 +90,7 @@ CGContextRef qt_mac_cg_context(const QPaintDevice *pdev)
 
 #endif
 
-class QQuickStyleNode : public QSGGeometryNode
+class QQuickStyleNode : public QSGNinePatchNode
 {
 public:
     QQuickStyleNode()
@@ -103,6 +98,8 @@ public:
     {
         m_geometry.setDrawingMode(GL_TRIANGLE_STRIP);
         setGeometry(&m_geometry);
+        // The texture material has mipmap filtering set to Nearest by default. This is not ideal.
+        m_material.setMipmapFiltering(QSGTexture::None);
         setMaterial(&m_material);
     }
 
@@ -111,38 +108,58 @@ public:
         delete m_material.texture();
     }
 
-    void initialize(QSGTexture *texture,
-                    const QRectF &bounds, qreal devicePixelRatio,
-                    int left, int top, int right, int bottom) {
-
+    virtual void setTexture(QSGTexture *texture)
+    {
         delete m_material.texture();
         m_material.setTexture(texture);
+    }
 
-        if (left <= 0 && top <= 0 && right <= 0 && bottom <= 0) {
+    virtual void setBounds(const QRectF &bounds)
+    {
+        m_bounds = bounds;
+    }
+
+    virtual void setDevicePixelRatio(qreal ratio)
+    {
+        m_devicePixelRatio = ratio;
+    }
+
+    virtual void setPadding(qreal left, qreal top, qreal right, qreal bottom)
+    {
+        m_paddingLeft = left;
+        m_paddingTop = top;
+        m_paddingRight = right;
+        m_paddingBottom = bottom;
+    }
+
+    virtual void update() {
+        QSGTexture *texture = m_material.texture();
+
+        if (m_paddingLeft <= 0 && m_paddingTop <= 0 && m_paddingRight <= 0 && m_paddingBottom <= 0) {
             m_geometry.allocate(4, 0);
-            QSGGeometry::updateTexturedRectGeometry(&m_geometry, bounds, texture->normalizedTextureSubRect());
+            QSGGeometry::updateTexturedRectGeometry(&m_geometry, m_bounds, texture->normalizedTextureSubRect());
             markDirty(QSGNode::DirtyGeometry | QSGNode::DirtyMaterial);
             return;
         }
 
         QRectF tc = texture->normalizedTextureSubRect();
         QSize ts = texture->textureSize();
-        ts.setHeight(ts.height() / devicePixelRatio);
-        ts.setWidth(ts.width() / devicePixelRatio);
+        ts.setHeight(ts.height() / m_devicePixelRatio);
+        ts.setWidth(ts.width() / m_devicePixelRatio);
 
         qreal invtw = tc.width() / ts.width();
         qreal invth = tc.height() / ts.height();
 
         struct Coord { qreal p; qreal t; };
-        Coord cx[4] = { { bounds.left(), tc.left() },
-                        { bounds.left() + left, tc.left() + left * invtw },
-                        { bounds.right() - right, tc.right() - right * invtw },
-                        { bounds.right(), tc.right() }
+        Coord cx[4] = { { m_bounds.left(), tc.left() },
+                        { m_bounds.left() + m_paddingLeft, tc.left() + m_paddingLeft * invtw },
+                        { m_bounds.right() - m_paddingRight, tc.right() - m_paddingRight * invtw },
+                        { m_bounds.right(), tc.right() }
                       };
-        Coord cy[4] = { { bounds.top(), tc.top() },
-                        { bounds.top() + top, tc.top() + top * invth },
-                        { bounds.bottom() - bottom, tc.bottom() - bottom * invth },
-                        { bounds.bottom(), tc.bottom() }
+        Coord cy[4] = { { m_bounds.top(), tc.top() },
+                        { m_bounds.top() + m_paddingTop, tc.top() + m_paddingTop * invth },
+                        { m_bounds.bottom() - m_paddingBottom, tc.bottom() - m_paddingBottom * invth },
+                        { m_bounds.bottom(), tc.bottom() }
                       };
 
         m_geometry.allocate(16, 28);
@@ -179,6 +196,12 @@ public:
 
     }
 
+    QRectF m_bounds;
+    qreal m_devicePixelRatio;
+    qreal m_paddingLeft;
+    qreal m_paddingTop;
+    qreal m_paddingRight;
+    qreal m_paddingBottom;
     QSGGeometry m_geometry;
     QSGTextureMaterial m_material;
 };
@@ -263,6 +286,8 @@ void QQuickStyleItem::initStyleOption()
                                         (sizeHint == "small") ? QPlatformTheme::SmallFont :
                                         QPlatformTheme::SystemFont;
 
+    bool needsResolvePalette = true;
+
     switch (m_itemType) {
     case Button: {
         if (!m_styleoption)
@@ -329,6 +354,10 @@ void QQuickStyleItem::initStyleOption()
         opt->features = QStyleOptionViewItem::HasDisplay;
         opt->text = text();
         opt->textElideMode = Qt::ElideRight;
+        opt->displayAlignment = Qt::AlignLeft | Qt::AlignVCenter;
+        opt->decorationAlignment = Qt::AlignCenter;
+        resolvePalette();
+        needsResolvePalette = false;
         QPalette pal = m_styleoption->palette;
         pal.setBrush(QPalette::Base, Qt::NoBrush);
         m_styleoption->palette = pal;
@@ -344,6 +373,7 @@ void QQuickStyleItem::initStyleOption()
 
         QStyleOptionHeader *opt = qstyleoption_cast<QStyleOptionHeader*>(m_styleoption);
         opt->text = text();
+        opt->textAlignment = static_cast<Qt::AlignmentFlag>(m_properties.value("textalignment").toInt());
         opt->sortIndicator = activeControl() == "down" ?
                     QStyleOptionHeader::SortDown
                   : activeControl() == "up" ?
@@ -503,19 +533,23 @@ void QQuickStyleItem::initStyleOption()
         // For GTK style. See below, in setElementType()
         setProperty("_q_isComboBoxPopupItem", m_itemType == ComboBoxItem);
 
-        QString scrollerDirection = m_properties["scrollerDirection"].toString();
-        if (!scrollerDirection.isEmpty()) {
+        QQuickMenuItemType::MenuItemType type =
+                static_cast<QQuickMenuItemType::MenuItemType>(m_properties["type"].toInt());
+        if (type == QQuickMenuItemType::ScrollIndicator) {
+            int scrollerDirection = m_properties["scrollerDirection"].toInt();
             opt->menuItemType = QStyleOptionMenuItem::Scroller;
-            opt->state |= scrollerDirection == "up" ?
+            opt->state |= scrollerDirection == Qt::UpArrow ?
                         QStyle::State_UpArrow : QStyle::State_DownArrow;
-        } else if (text().isEmpty()) {
+        } else if (type == QQuickMenuItemType::Separator) {
             opt->menuItemType = QStyleOptionMenuItem::Separator;
         } else {
             opt->text = text();
 
-            if (m_properties["isSubmenu"].toBool()) {
+            if (type == QQuickMenuItemType::Menu) {
                 opt->menuItemType = QStyleOptionMenuItem::SubMenu;
             } else {
+                opt->menuItemType = QStyleOptionMenuItem::Normal;
+
                 QString shortcut = m_properties["shortcut"].toString();
                 if (!shortcut.isEmpty()) {
                     opt->text += QLatin1Char('\t') + shortcut;
@@ -527,8 +561,6 @@ void QQuickStyleItem::initStyleOption()
                     QVariant exclusive = m_properties["exclusive"];
                     opt->checkType = exclusive.toBool() ? QStyleOptionMenuItem::Exclusive :
                                                           QStyleOptionMenuItem::NonExclusive;
-                } else {
-                    opt->menuItemType = QStyleOptionMenuItem::Normal;
                 }
             }
             if (m_properties["icon"].canConvert<QIcon>())
@@ -601,6 +633,7 @@ void QQuickStyleItem::initStyleOption()
 
         QStyleOptionSpinBox *opt = qstyleoption_cast<QStyleOptionSpinBox*>(m_styleoption);
         opt->frame = true;
+        opt->subControls = QStyle::SC_SpinBoxFrame | QStyle::SC_SpinBoxEditField;
         if (value() & 0x1)
             opt->activeSubControls = QStyle::SC_SpinBoxUp;
         else if (value() & (1<<1))
@@ -710,6 +743,9 @@ void QQuickStyleItem::initStyleOption()
     if (!m_styleoption)
         m_styleoption = new QStyleOption();
 
+    if (needsResolvePalette)
+        resolvePalette();
+
     m_styleoption->styleObject = this;
     m_styleoption->direction = qApp->layoutDirection();
 
@@ -743,6 +779,15 @@ void QQuickStyleItem::initStyleOption()
     if (m_horizontal)
         m_styleoption->state |= QStyle::State_Horizontal;
 
+    // some styles don't draw a focus rectangle if
+    // QStyle::State_KeyboardFocusChange is not set
+    if (window()) {
+         Qt::FocusReason lastFocusReason = QQuickWindowPrivate::get(window())->lastFocusReason;
+         if (lastFocusReason == Qt::TabFocusReason || lastFocusReason == Qt::BacktabFocusReason) {
+             m_styleoption->state |= QStyle::State_KeyboardFocusChange;
+         }
+    }
+
     if (sizeHint == "mini") {
         m_styleoption->state |= QStyle::State_Mini;
     } else if (sizeHint == "small") {
@@ -751,6 +796,62 @@ void QQuickStyleItem::initStyleOption()
 
 }
 
+void QQuickStyleItem::resolvePalette()
+{
+    QPlatformTheme::Palette paletteType = QPlatformTheme::SystemPalette;
+    switch (m_itemType) {
+    case Button:
+        paletteType = QPlatformTheme::ButtonPalette;
+        break;
+    case RadioButton:
+        paletteType = QPlatformTheme::RadioButtonPalette;
+        break;
+    case CheckBox:
+        paletteType = QPlatformTheme::CheckBoxPalette;
+        break;
+    case ComboBox:
+    case ComboBoxItem:
+        paletteType = QPlatformTheme::ComboBoxPalette;
+        break;
+    case ToolBar:
+    case ToolButton:
+        paletteType = QPlatformTheme::ToolButtonPalette;
+        break;
+    case Tab:
+    case TabFrame:
+        paletteType = QPlatformTheme::TabBarPalette;
+        break;
+    case Edit:
+        paletteType = QPlatformTheme::TextEditPalette;
+        break;
+    case GroupBox:
+        paletteType = QPlatformTheme::GroupBoxPalette;
+        break;
+    case Header:
+        paletteType = QPlatformTheme::HeaderPalette;
+        break;
+    case Item:
+    case ItemRow:
+        paletteType = QPlatformTheme::ItemViewPalette;
+        break;
+    case Menu:
+    case MenuItem:
+        paletteType = QPlatformTheme::MenuPalette;
+        break;
+    case MenuBar:
+    case MenuBarItem:
+        paletteType = QPlatformTheme::MenuBarPalette;
+        break;
+    default:
+        break;
+    }
+
+    const QPalette *platformPalette = QGuiApplicationPrivate::platformTheme()->palette(paletteType);
+    if (platformPalette)
+        m_styleoption->palette = *platformPalette;
+    // Defaults to SystemPalette otherwise
+}
+
 /*
  *   Property style
  *
@@ -866,8 +967,16 @@ QSize QQuickStyleItem::sizeFromContents(int width, int height)
         break;
     case Button: {
         QStyleOptionButton *btn = qstyleoption_cast<QStyleOptionButton*>(m_styleoption);
-        int newWidth = qMax(width, btn->fontMetrics.width(btn->text));
-        int newHeight = qMax(height, btn->fontMetrics.height());
+
+        int contentWidth = btn->fontMetrics.width(btn->text);
+        int contentHeight = btn->fontMetrics.height();
+        if (!btn->icon.isNull()) {
+            //+4 matches a hardcoded value in QStyle and acts as a margin between the icon and the text.
+            contentWidth += btn->iconSize.width() + 4;
+            contentHeight = qMax(btn->fontMetrics.height(), btn->iconSize.height());
+        }
+        int newWidth = qMax(width, contentWidth);
+        int newHeight = qMax(height, contentHeight);
         size = qApp->style()->sizeFromContents(QStyle::CT_PushButton, m_styleoption, QSize(newWidth, newHeight)); }
 #ifdef Q_OS_OSX
         if (style() == "mac") {
@@ -897,7 +1006,7 @@ QSize QQuickStyleItem::sizeFromContents(int width, int height)
             size = qApp->style()->sizeFromContents(QStyle::CT_SpinBox, m_styleoption, QSize(width, height + 5));
             break;
         }
-#endif // fall trough if not mac
+#endif // fall through if not mac
     case Edit:
 #ifdef Q_OS_OSX
         if (style() =="mac") {
@@ -919,6 +1028,9 @@ QSize QQuickStyleItem::sizeFromContents(int width, int height)
             frame.rect = m_styleoption->rect;
             frame.styleObject = this;
             size = qApp->style()->sizeFromContents(QStyle::CT_LineEdit, &frame, QSize(width, height));
+            if (m_itemType == SpinBox)
+                size.setWidth(qApp->style()->sizeFromContents(QStyle::CT_SpinBox,
+                                                              m_styleoption, QSize(width + 2, height)).width());
         }
         break;
     case GroupBox: {
@@ -971,6 +1083,7 @@ QSize QQuickStyleItem::sizeFromContents(int width, int height)
 qreal QQuickStyleItem::baselineOffset()
 {
     QRect r;
+    bool ceilResult = true; // By default baseline offset rounding is done upwards
     switch (m_itemType) {
     case RadioButton:
         r = qApp->style()->subElementRect(QStyle::SE_RadioButtonContents, m_styleoption);
@@ -992,16 +1105,20 @@ qreal QQuickStyleItem::baselineOffset()
         }
         break;
     case SpinBox:
-        if (const QStyleOptionSpinBox *spinbox = qstyleoption_cast<const QStyleOptionSpinBox *>(m_styleoption))
+        if (const QStyleOptionSpinBox *spinbox = qstyleoption_cast<const QStyleOptionSpinBox *>(m_styleoption)) {
             r = qApp->style()->subControlRect(QStyle::CC_SpinBox, spinbox, QStyle::SC_SpinBoxEditField);
+            ceilResult = false;
+        }
         break;
     default:
         break;
     }
-    if (r.isValid()) {
+    if (r.height() > 0) {
         const QFontMetrics &fm = m_styleoption->fontMetrics;
-        const float surplus = r.height() - fm.height();
-        float result = float(r.top()) + surplus/2.0 + fm.ascent();
+        int surplus = r.height() - fm.height();
+        if ((surplus & 1) && ceilResult)
+            surplus++;
+        int result = r.top() + surplus/2 + fm.ascent();
 #ifdef Q_OS_OSX
         if (style() == QStringLiteral("mac")) {
             switch (m_itemType) {
@@ -1066,7 +1183,7 @@ int QQuickStyleItem::pixelMetric(const QString &metric)
     if (metric == "scrollbarExtent")
         return qApp->style()->pixelMetric(QStyle::PM_ScrollBarExtent, 0 );
     else if (metric == "defaultframewidth")
-        return qApp->style()->pixelMetric(QStyle::PM_DefaultFrameWidth, 0);
+        return qApp->style()->pixelMetric(QStyle::PM_DefaultFrameWidth, m_styleoption);
     else if (metric == "taboverlap")
         return qApp->style()->pixelMetric(QStyle::PM_TabBarTabOverlap, 0 );
     else if (metric == "tabbaseoverlap")
@@ -1112,11 +1229,9 @@ QVariant QQuickStyleItem::styleHint(const QString &metric)
     if (metric == "comboboxpopup") {
         return qApp->style()->styleHint(QStyle::SH_ComboBox_Popup, m_styleoption);
     } else if (metric == "highlightedTextColor") {
-        QPalette pal = QApplication::palette("QAbstractItemView");
-        pal.setCurrentColorGroup(m_styleoption->palette.currentColorGroup());
-        return pal.highlightedText().color().name();
+        return m_styleoption->palette.highlightedText().color().name();
     } else if (metric == "textColor") {
-        QPalette pal = qApp->palette();
+        QPalette pal = m_styleoption->palette;
         pal.setCurrentColorGroup(active()? QPalette::Active : QPalette::Inactive);
         return pal.text().color().name();
     } else if (metric == "focuswidget") {
@@ -1132,9 +1247,11 @@ QVariant QQuickStyleItem::styleHint(const QString &metric)
         return qApp->style()->styleHint(QStyle::SH_ScrollBar_LeftClickAbsolutePosition);
     else if (metric == "activateItemOnSingleClick")
         return qApp->style()->styleHint(QStyle::SH_ItemView_ActivateItemOnSingleClick);
+    else if (metric == "submenupopupdelay")
+        return qApp->style()->styleHint(QStyle::SH_Menu_SubMenuPopupDelay, m_styleoption);
     return 0;
 
-    // Add SH_Menu_SpaceActivatesItem, SH_Menu_SubMenuPopupDelay
+    // Add SH_Menu_SpaceActivatesItem
 }
 
 void QQuickStyleItem::setHints(const QVariantMap &str)
@@ -1521,6 +1638,15 @@ void QQuickStyleItem::paint(QPainter *painter)
         painter->setPen(pen);
     }    break;
     case SpinBox:
+#ifdef Q_OS_MAC
+        // macstyle depends on the embedded qlineedit to fill the editfield background
+        if (style() == "mac") {
+            QRect editRect = qApp->style()->subControlRect(QStyle::CC_SpinBox,
+                                                           qstyleoption_cast<QStyleOptionComplex*>(m_styleoption),
+                                                           QStyle::SC_SpinBoxEditField);
+            painter->fillRect(editRect.adjusted(-1, -1, 1, 1), m_styleoption->palette.base());
+        }
+#endif
         qApp->style()->drawComplexControl(QStyle::CC_SpinBox,
                                           qstyleoption_cast<QStyleOptionComplex*>(m_styleoption),
                                           painter);
@@ -1584,6 +1710,7 @@ void QQuickStyleItem::paint(QPainter *painter)
             frame.midLineWidth = 0;
             frame.rect = m_styleoption->rect;
             frame.styleObject = this;
+            frame.palette = m_styleoption->palette;
             qApp->style()->drawPrimitive(QStyle::PE_FrameMenu, &frame, painter);
         }
     }
@@ -1656,9 +1783,12 @@ QSGNode *QQuickStyleItem::updatePaintNode(QSGNode *node, UpdatePaintNodeData *)
         return 0;
     }
 
-    QQuickStyleNode *styleNode = static_cast<QQuickStyleNode *>(node);
-    if (!styleNode)
-        styleNode = new QQuickStyleNode;
+    QSGNinePatchNode *styleNode = static_cast<QSGNinePatchNode *>(node);
+    if (!styleNode) {
+        styleNode = QQuickItemPrivate::get(this)->sceneGraphContext()->createNinePatchNode();
+        if (!styleNode)
+            styleNode = new QQuickStyleNode;
+    }
 
 #ifdef QSG_RUNTIME_DESCRIPTION
     qsgnode_set_description(styleNode,
@@ -1668,10 +1798,12 @@ QSGNode *QQuickStyleItem::updatePaintNode(QSGNode *node, UpdatePaintNodeData *)
                             .arg(text()));
 #endif
 
-    styleNode->initialize(window()->createTextureFromImage(m_image, QQuickWindow::TextureCanUseAtlas),
-                          boundingRect(),
-                          window()->devicePixelRatio(),
-                          m_border.left(), m_border.top(), m_border.right(), m_border.bottom());
+    styleNode->setTexture(window()->createTextureFromImage(m_image, QQuickWindow::TextureCanUseAtlas));
+    styleNode->setBounds(boundingRect());
+    styleNode->setDevicePixelRatio(window()->devicePixelRatio());
+    styleNode->setPadding(m_border.left(), m_border.top(), m_border.right(), m_border.bottom());
+    styleNode->update();
+
     return styleNode;
 }
 
